'*************************************************************************************************************
'* This source code is provided 'as is' and any express or implied warranties, including,                    *
'* but not limited to, the implied warranties of merchantability and fitness for a particular                *
'* purpose are disclaimed.                                                                                   *
'*                                                                                                           *
'* In no event shall the author, Crownhill or contributers be liable for any direct, indirect, incidental,   *
'* special, exemplary, or consequential damages (including, but not limited to,  procurement of substitute   *
'* goods or services; loss of use, data, or profits;  or business interruption) however caused and on any    *
'* theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise)   *
'* arising in any way out of the use of this source code, even if advised of the possibility of such damage. *
'*                                                                                                           *
'* Any views and conclusions contained in the source code and any documentation are those of the author's    *
'* and should not be interpreted as representing official policies,                                          *
'* either expressed or implied, of Crownhill Associates Ltd.                                                 *
'*************************************************************************************************************
'
' Pulse rate capture using a compare special event and INT0 interrupt
'
' Uses a Dword for the time gate window so it can be a very long period indeed. Days in fact
' Also uses a Dword for the pulse total
'
' This version stores the total pulses within internal eeprom
' It also has a serial interface to read or clear the pulses
'
' The command: read<CR> with return the current pulse total
' The command: reset<CR> will reset the pulse total
'
' Written by Les Johnson for the Proton compiler version 3.5.5.4 using the Amicus18 hardware platform
'
    #Disable HRSIn, HRSInTo, HRSIn_RCREG_Read, HRSOut       ' Disable the library subroutines for HRSOut and HRSIn(with and without timeouts)
    
    Include "Amicus18.inc"                          ' Configure the compiler to use the Amicus18 hardware (18F25K20, 64MHz)
    Include "Amicus18_Timers.inc"                   ' Include the Amicus18 Timer macros into the program
'
' Setup the LCD
'    
    Declare LCD_DTPin = PORTB.4                     ' LCD's data port starts at bit-4 of PORTB
    Declare LCD_ENPin = PORTB.3                     ' LCD's EN pin connection
    Declare LCD_RSPin = PORTB.2                     ' LCD's RS pin connection
    Declare LCD_Interface = 4                       ' LCD operating in 4-bit mode
    Declare LCD_Lines = 2                           ' LCD has 2 lines
    Declare LCD_Type = Alphanumeric                 ' LCD type is alphanumeric (Hitachi)
    Declare LCD_CommandUs = 2000                    ' Delay between command values sent to the LCD (in microseconds)
    Declare LCD_DataUs = 50                         ' Delay between data values sent to the LCD (in microseconds) 
    
    On_Hardware_Interrupt GoTo ISR_Handler          ' Point to the interrupt handler

    Dim bCapStatusFlags As Byte System              ' Keep interrupt flag in Access RAM
    Dim tCaptureComplete As bCapStatusFlags.0       ' True if a capture has occurred (must be reset in the main program)
    Dim tWriteResult As bCapStatusFlags.1           ' True if a write to external memory was accomplished correctly
'
' Pulse counting variables
'    
    Dim wIntPulseRate As Word System                ' Pulse rate counter used within the interrupt
    Dim dPulseTotal As Dword System                 ' Pulse rate total          
    Dim dMilliSeconds As Dword System               ' Millisecond counter used for a capture window
'
' Eeprom variables
'    
    Dim wEepromAddress As Word                      ' Current address within the internal eeprom
'
' Create the variables for the interrupt driven USART1 Tx and Rx routines 
'
    Dim USART_bSerTemp As Byte System                       ' Temporary variable used in the interrupt routine
'
' Any variable used in the interrupt should to be in Access RAM for non-banked operation
'   
    Dim USART1_bRxInIndex As Byte System                    ' \
    Dim USART1_bRxOutIndex As Byte System                   ' / USART1 Rx buffer index variables    
    Dim USART1_bTxInIndex As Byte System                    ' \
    Dim USART1_bTxOutIndex As Byte System                   ' / USART1 Tx buffer index variables
'
' Create the Tx and Rx buffers for USART1
'    
    Symbol cUSART1_RxBufferSize = 32                        ' Rx Serial buffer size must be a power of 2. i.e. 2, 4, 8, 16, 32, 64 or 128
    Symbol cUSART1_RxFIFO_Mask = cUSART1_RxBufferSize - 1   ' Create a mask for resetting or testing the serial buffers
    Dim USART1_bRxBuffer[cUSART1_RxBufferSize] As Byte Heap ' USART1 Rx circular buffer
    
    Symbol cUSART1_TxBufferSize = 32                        ' Tx Serial buffer size must be a power of 2. i.e. 2, 4, 8, 16, 32, 64 or 128
    Symbol cUSART1_TxFIFO_Mask = cUSART1_TxBufferSize - 1   ' Create a mask for resetting or testing the serial buffers
    Dim USART1_bTxBuffer[cUSART1_TxBufferSize] As Byte Heap ' USART1 Tx circular buffer
'
' Create some compiler System variables
'
    Dim GEN As Byte System                                  ' \
    Dim GENH As Byte System                                 ' / Timeout value storage
    Dim PP0 As Byte System                                  ' Used by the HRSIn and HRSOut routines
    Dim PP4 As Byte System                                  ' Used by Hbusout for Ack response testing
    
    Dim USART_bTimeoutInt As GEN
    Dim USART_bTimeoutIntH As GENH
    Dim USART_wTimeoutValue As USART_bTimeoutInt.Word       ' Alias the timeout value to GEN\H   
    Dim USART_wInsideLoop As Word System                    ' Inside loop for HRSIn with timeout
    Dim USART_wOutsideLoop As Word System                   ' Outside loop for HRSIn with timeout
'
' Create the variables used in the main program loop
'    
    Dim CommandString As String * 20                        ' Holds a command received serially
    
$ifndef False
    $define False 0
$endif
$ifndef True
    $define True 1
$endif
    Dim wCCPR2 As  CCPR2L .Word                             ' Combine CCPR2L\H into a 16-bit register
    Dim USART_wPROD As PRODL.Word                           ' Combine PRODL\H into a 16-bit register   
    Dim USART_wFSR0 As FSR0L.Word                           ' Combine FSR0L\H into a 16-bit register
'--------------------------------------------------------------------------------- 
'
' Set the pulse total count storage within internal eeprom
'
PulseStorage EData Dword 0
      
'---------------------------------------------------------------------------------    
    GoTo Main                                               ' Jump over the subroutines 
'---------------------------------------------------------------------------------
' Interrupt handler
' Operate a time window for pulse samples
' Count pulses using an INT0 event
'
' Input     : None
' Output    : Variable dPulseTotal holds the total pulse count
'           : tCaptureComplete is true if a frame has been completed (must be cleared in the main program)
' Notes     : Change the value compared with dMilliSeconds to alter the time gate period
'
ISR_Handler:
    Context Save
'
' Service a special event interrupt every 10 milliseconds
'    
    If PIR2bits_CCP2IF = 1 Then                             ' Was it a Compare Special event on CCP2 that triggered the interrupt?       
        Inc dMilliSeconds                                   ' Yes. So increment the milliseconds counter
        If dMilliSeconds >= 100 Then                        ' Have we reached the end of the time window?
            Clear dMilliSeconds                             ' Yes. So clear the millisecond counter
            dPulseTotal = dPulseTotal + wIntPulseRate       ' Add the pulse count to the total
            Clear wIntPulseRate                             ' Reset the interrupt pulse count variable
            tCaptureComplete = True                         ' Signal that a pulse rate capture is complete (must be Reset in the Main program)
        EndIf
        Clear PIR2bits_CCP2IF                               ' Clear the interrupt flag
    EndIf
'
' Service an INT0 external interrupt
'
    If INTCONbits_INT0IF = 1 Then                           ' Was it an INT0 event that triggered the interrupt?      
        Inc wIntPulseRate                                  ' Yes. So increment the pulse count
        Clear INTCONbits_INT0IF                             ' Clear the INT0 flag
    EndIf
'
' Service a USART1 Rx interrupt
'
    If PIR1bits_RCIF = 1 Then                               ' Was it a USART1 receive that triggered the interrupt?                            
        If RCSTAbits_OERR = 1 Then                          ' Yes. So is there an error in USART1 reception?
            WREG = RCREG                                    ' \ Yes. So empty the USART1 Rx buffer
            WREG = RCREG                                    ' /
            Clear RCSTAbits_CREN                            ' Disable USART1 Rx
            Set RCSTAbits_CREN                              ' Re-enable again to clear errors
        Else                                                ' Otherwise. No error so...
            USART1_bRxBuffer[USART1_bRxInIndex] = RCREG     ' Place the byte received into the USART1 buffer      
            WREG = USART1_bRxInIndex + 1                    ' Increment the index in WREG 
            USART_bSerTemp = WREG & cUSART1_RxFIFO_Mask     ' Reset the USART1 buffer when full
            If USART_bSerTemp <> USART1_bRxOutIndex Then    ' Have we reached the end of the USART1 buffer?               
                USART1_bRxInIndex = USART_bSerTemp          ' No. So move to the next position
            EndIf
        EndIf
    EndIf
'
' Service a USART1 Tx interrupt
'                                       
    If PIR1bits_TXIF = 1 Then                               ' Was it a USART1 transmit that triggered the interrupt? 
        If PIE1bits_TXIE = 1 Then                           ' Yes. So are USART1 Tx interrupts enabled? 
            TXREG = USART1_bTxBuffer[USART1_bTxOutIndex]    ' Yes. So transmit a byte from the USART1 buffer      
            Inc USART1_bTxOutIndex                          ' Increment the out USART1 buffer Index 
            USART1_bTxOutIndex = USART1_bTxOutIndex & cUSART1_TxFIFO_Mask ' Reset when full            
            If USART1_bTxOutIndex = USART1_bTxInIndex Then  ' Are there any more bytes in the USART1 buffer?             
                Clear PIE1bits_TXIE                         ' No. So disable USART1 Tx interrupts 
            EndIf                       
        EndIf
    EndIf
'
' Other interrupt servicing goes here
'
     Context Restore                                        ' Exit the interrupt    
'-------------------------------------------------------------------------------------------
' Wait for a byte from the interrupt driven circular USART1 buffer with timeout
' Input     : GEN\GENH hold the timeout value in approx ms (0 to 65535)
' Output    : PP0 and WREG hold the value received
'           : Carry Flag (STATUS.0) Clear if timeout occured
' Notes     : Replacement for the compiler's HRSIn command with timeout
'
#ifdef HRSInTo#Req
HRSIn2:
    USART_wOutsideLoop = USART_wTimeoutValue                ' Save the timeout value so it doesn't get overwritten
    Clear USART_wInsideLoop                                 ' Reset the inside loop counter
USART1_GetbufOutLoop:
    DelayCS 2                                               ' Delay for 2 cycles within the outside loop
USART1_GetbufInLoop:
    DelayCS 1                                               ' Delay for 1 cycle within the inside loop
    WREG = USART1_bRxOutIndex - USART1_bRxInIndex           ' Is there anything in the serial buffer?
    Bnz USART1_Getch                                        ' Yes. So fetch it
    WREG = 255                                              ' \
    Addwf  USART_wInsideLoop,f                              ' |
    Addwfc USART_wInsideLoopH,f                             ' | Decrement the inside and outside loops
    Addwfc USART_wOutsideLoop,f                             ' |
    Addwfc USART_wOutsideLoopH,f                            ' /
    Btfss STATUS,0
    Ret                                                     ' Return with the Carry flag clear to indicate timed out
    Infsnz USART_wInsideLoop,w
    Incfsz USART_wInsideLoopH,w
    Bra USART1_GetbufOutLoop
    USART_wInsideLoop = ((59 * Xtal) / 4)                   ' Set the inside loop counter based upon the XTAL frequency
    Bra USART1_GetbufInLoop
#endif  ' HRSInTo#Req       
'-------------------------------------------------------------------------------------------
' Retrieve a byte from the USART1 Rx buffer  
' Input     : None
' Output    : PP0 and WREG hold the oldest byte from the buffer
' Notes     : Waits for a byte if none present (blocking)
'           : Replacement for the compiler's HRSIn command without timeout
'
#ifdef HRSInX#Req
@HRSIn
    Repeat                                                  ' \
    #ifdef WatchDog_Req                                     ' |
        Clrwdt                                              ' | Clear the watchdog timer within the loop if required
    #endif                                                  ' |
    Until USART1_bRxInIndex <> USART1_bRxOutIndex           ' / Wait until a character is present in the buffer
'
' Get a byte from the buffer (non-blocking)
'
USART1_Getch:    
    Clear PIE1bits_RCIE                                     ' Disable USART1 Rx interrupts
    PP0 = USART1_bRxBuffer[USART1_bRxOutIndex]              ' Get the oldest byte received from the USART1 buffer
    Inc USART1_bRxOutIndex                                  ' Move the Index to discard USART1 buffer
    USART1_bRxOutIndex = USART1_bRxOutIndex & cUSART1_RxFIFO_Mask ' If the Index is at end, roll-it over.
    Set PIE1bits_RCIE                                       ' Enable USART1 Rx interrupts
    WREG = PP0                                              ' Place the byte received into WREG
    Set STATUSbits_C                                        ' Set the Carry flag to indicate a byte received
    Return
#endif  ' HRSInX#Req    
'-------------------------------------------------------------------------------------------
' Transmit a single byte from the buffer to USART1
' Input     : WREG holds the byte to transmit
' Output    : WREG still holds the byte transmitted
' Notes     : If the buffer is full, wait until an empty slot is present
'           : Replacement for the compiler's HRSOut command
'
#ifdef HRSOut#Req
@HRSOut
    PP0 = WREG                                              ' Transfer WREG to its working variable
    USART_wPROD = USART_wFSR0                               ' Save FSR0L\H
    #ifdef WatchDog_Req                                     ' \
        Clrwdt                                              ' | Clear the watchdog timer if required
    #endif                                                  ' /
    Repeat                                                  ' \
        WREG = USART1_bTxInIndex + 1                        ' |
        WREG = WREG & cUSART1_TxFIFO_Mask                   ' | Wait until the USART1 Tx buffer has an empty slot
    Until WREG <> USART1_bTxOutIndex                        ' /      
    Clear PIE1bits_TXIE                                     ' Disable the USART1 Tx interrupt
    USART1_bTxBuffer[USART1_bTxInIndex] = PP0               ' Place a byte to transmit into the buffer
    Inc USART1_bTxInIndex                                   ' Increment the In buffer Index
    USART1_bTxInIndex = USART1_bTxInIndex & cUSART1_TxFIFO_Mask ' Reset when full 
    Set PIE1bits_TXIE                                       ' Enable the USART1 Tx Interrupt
    USART_wFSR0 = USART_wPROD                               ' Restore FSR0L\H
    WREG = PP0                                              ' Restore WREG before leaving
    Return
#endif  ' HRSOut#Req

'-------------------------------------------------------------------------------------------
' Initialise the serial buffers and interrupts
' Input     : None
' Output    : None
' Notes     : None
'
USARTx_Init:
    Clear PIE1bits_TXIE                                     ' Disable USART1 Tx interrupts
    Set PIE1bits_RCIE                                       ' Enable USART1 Rx interrupts
    Clear USART1_bRxInIndex                                 ' \
    Clear USART1_bRxOutIndex                                ' / Clear USART1 Rx buffer Indexes
    Clear USART1_bTxInIndex                                 ' \
    Clear USART1_bTxOutIndex                                ' / Clear USART1 Tx buffer Indexes

    Set INTCONbits_PEIE                                     ' Enable peripheral interrupts
    Set INTCONbits_GIE                                      ' Enable global interrupts
    Return
'----------------------------------------------------------------------------
' Write the amount of pulses to internal eeprom
' Input     : dPulseTotal holds the 32-bit running total of pulses
' Output    : None
' Notes     : In reality, the eeprom address will need to be cycled
'           : Otherwise, it may wear out.
' 
WriteTotalToEeprom:
    tWriteResult = True                                 ' Default to a correct write
    Clear INTCONbits_GIE                                ' Disable global interrupts
    EWrite wEepromAddress,[dPulseTotal]                 ' Write the pulse total to internal eeprom
    Set INTCONbits_GIE                                  ' Enable global interrupts
    Return

'----------------------------------------------------------------------------
' Initialise variables, flags and the interrupts
' Input     : None
' Output    : None
' Notes     : None
'
Initialise:
    Clear wEepromAddress                                ' Reset the eeprom address
    Clear dMilliSeconds                                 ' Clear the millisecond counter
    Clear wIntPulseRate                                 ' Clear the internal pulse rate counter
    dPulseTotal = ERead wEepromAddress                  ' Read the last pulse total from internal eeprom   
    tCaptureComplete = False                            ' Clear the capture event flag
    CCP2CON = $0B                                       ' Compare mode: trigger special event, reset timer, start A/D conversion on CCP2 match (tCCP2IF bit is set)
'
' Configure Timer3
'
    OpenTimer3(TIMER_INT_OFF & T3_16BIT_RW & T3_PS_1_8 & T3_SYNC_EXT_OFF & T3_SOURCE_INT & T3_SOURCE_CCP)
    $define cPrescalerValue 8                           ' Alter this to match the prescaler parameter above. i.e. 4 for T3_PS_1_4, 8 for T3_PS_1_8
    $define cMicroSeconds 10000                         ' Interrupt rate (in uS)
'
' Calculate the value to place into the CCPRx register in order to achieve a certain interrupt rate (in us)
'
    $define cCCPR_Value $eval (cMicroSeconds / cPrescalerValue) * (_xtal / 4)

$if cCCPR_Value > 65535
    $error "Value too large for interrupt duration"
$elseif cCCPR_Value = 0
    $error "Value too small for interrupt duration"
$endif

    wCCPR2 = cCCPR_Value                                ' Load CCPR2L\H with the value to trigger an interrupt at a certain duration
    Set PIE2bits_CCP2IE                                 ' Enable the Special Event Interrupt on CCP2
    
    Input PORTB.0                                       ' Make sure the INT0 pin is an input
    Set INTCON2bits_INTEDG0                             ' INT0 triggers on a rising edge
    Set INTCONbits_INT0IE                               ' Enable an INT0 interrupt
    
    GoSub USARTx_Init                                   ' Initialise the interrupt driven serial routines
    'Set INTCONbits_PEIE                                 ' Enable peripheral interrupts
    'Set INTCONbits_GIE                                  ' Enable global interrupts
    Return
    
'----------------------------------------------------------------------------
' The main program loop starts here
'    
Main:
    DelayMS 10                                          ' Wait for things to stabilise
    Cls                                                 ' Clear the LCD    
    GoSub Initialise                                    ' Initialise the program    
    
    Print At 1, 1, "Pulses ", Dec dPulseTotal           ' Display the default pulses text
    HRSOut ">"                                          ' Add a serial prompt
'
' Create a loop to display the pulse rate total and store within internal eeprom
' Receive serial commands from USART1 and act appropriately
'       
    While 1 = 1                                         ' Create an infinite loop
        If tCaptureComplete = True Then                 ' Is there a pulse rate ready?
            tCaptureComplete = False                    ' Yes. So Reset the capture event flag
            Print At 1, 1, "Pulses ", Dec dPulseTotal   ' Display the total amount of pulses
            GoSub WriteTotalToEeprom                    ' Write the 32-bit total to internal eeprom
        EndIf   
        
        If USART1_bRxInIndex <> USART1_bRxOutIndex Then ' Have we received anything serially?     
            HRSIn {500,TimedOut},CommandString          ' Yes. So get it from the buffer (with a timeout)
            CommandString = ToLower(CommandString)      ' Convert it to lower case in order to make it easier to test
            If CommandString = "read" Then              ' Did we see a read command?
                HRSOut Dec dPulseTotal                  ' Yes. So transmit the total pulses
            ElseIf CommandString = "reset" Then         ' Did we see a reset command?
                Clear INTCONbits_GIE                    ' Yes. So disable interrupts for now
                Clear dMilliSeconds                     ' Clear the millisecond counter
                Clear dPulseTotal                       ' Clear the pulse total value
                Clear wIntPulseRate                     ' Clear the internal pulse count
                Set INTCONbits_GIE                      ' Re-Enable interrupts
                Print At 1, 1, "Pulses 0         "      ' Reset the line on the LCD
                HRSOut "OK"                             ' Transmit a comfort response
            Else                                        ' Otherwise...
                HRSOut "ERROR"                          ' Command error
            EndIf
TimedOut:
            HRSOut "\r>"                                ' Transmit a serial prompt    
        EndIf
    Wend                                                ' Do it forever
